@@ -41,17 +41,6 @@ static PyObject *from_array(MMDB_entry_data_list_s **entry_data_list);
 static PyObject *from_uint128(const MMDB_entry_data_list_s *entry_data_list);
 static int ip_converter(PyObject *obj, struct sockaddr_storage *ip_address);
 
-#if PY_MAJOR_VERSION >= 3
-#define MOD_INIT(name) PyMODINIT_FUNC PyInit_##name(void)
-#define RETURN_MOD_INIT(m) return (m)
-#define FILE_NOT_FOUND_ERROR PyExc_FileNotFoundError
-#else
-#define MOD_INIT(name) PyMODINIT_FUNC init##name(void)
-#define RETURN_MOD_INIT(m) return
-#define PyInt_FromLong PyLong_FromLong
-#define FILE_NOT_FOUND_ERROR PyExc_IOError
-#endif
-
 #ifdef __GNUC__
 #define UNUSED(x) UNUSED_##x __attribute__((__unused__))
 #else
@@ -59,16 +48,27 @@ static int ip_converter(PyObject *obj, struct sockaddr_storage *ip_address);
 #endif
 
 static int Reader_init(PyObject *self, PyObject *args, PyObject *kwds) {
-    char *filename;
+    PyObject *filepath = NULL;
     int mode = 0;
 
     static char *kwlist[] = {"database", "mode", NULL};
-    if (!PyArg_ParseTupleAndKeywords(
-            args, kwds, "s|i", kwlist, &filename, &mode)) {
+    if (!PyArg_ParseTupleAndKeywords(args,
+                                     kwds,
+                                     "O&|i",
+                                     kwlist,
+                                     PyUnicode_FSConverter,
+                                     &filepath,
+                                     &mode)) {
+        return -1;
+    }
+
+    char *filename = PyBytes_AS_STRING(filepath);
+    if (filename == NULL) {
         return -1;
     }
 
     if (mode != 0 && mode != 1) {
+        Py_XDECREF(filepath);
         PyErr_Format(
             PyExc_ValueError,
             "Unsupported open mode (%i). Only "
@@ -78,25 +78,29 @@ static int Reader_init(PyObject *self, PyObject *args, PyObject *kwds) {
     }
 
     if (0 != access(filename, R_OK)) {
-        PyErr_Format(
-            FILE_NOT_FOUND_ERROR, "No such file or directory: '%s'", filename);
+
+        PyErr_SetFromErrnoWithFilenameObject(PyExc_OSError, filepath);
+        Py_XDECREF(filepath);
         return -1;
     }
 
     MMDB_s *mmdb = (MMDB_s *)malloc(sizeof(MMDB_s));
     if (NULL == mmdb) {
+        Py_XDECREF(filepath);
         PyErr_NoMemory();
         return -1;
     }
 
     Reader_obj *mmdb_obj = (Reader_obj *)self;
     if (!mmdb_obj) {
+        Py_XDECREF(filepath);
         free(mmdb);
         PyErr_NoMemory();
         return -1;
     }
 
-    uint16_t status = MMDB_open(filename, MMDB_MODE_MMAP, mmdb);
+    int const status = MMDB_open(filename, MMDB_MODE_MMAP, mmdb);
+    Py_XDECREF(filepath);
 
     if (MMDB_SUCCESS != status) {
         free(mmdb);
@@ -127,7 +131,10 @@ static PyObject *Reader_get_with_prefix_len(PyObject *self, PyObject *args) {
         return NULL;
     }
 
-    return PyTuple_Pack(2, record, PyLong_FromLong(prefix_len));
+    PyObject *tuple = Py_BuildValue("(Oi)", record, prefix_len);
+    Py_DECREF(record);
+
+    return tuple;
 }
 
 static int get_record(PyObject *self, PyObject *args, PyObject **record) {
@@ -178,6 +185,7 @@ static int get_record(PyObject *self, PyObject *args, PyObject **record) {
     }
 
     if (!result.found_entry) {
+        Py_INCREF(Py_None);
         *record = Py_None;
         return prefix_len;
     }
@@ -200,21 +208,19 @@ static int get_record(PyObject *self, PyObject *args, PyObject **record) {
     *record = from_entry_data_list(&entry_data_list);
     MMDB_free_entry_data_list(original_entry_data_list);
 
+    // from_entry_data_list will return NULL on errors.
+    if (*record == NULL) {
+        return -1;
+    }
+
     return prefix_len;
 }
 
 static int ip_converter(PyObject *obj, struct sockaddr_storage *ip_address) {
-#if PY_MAJOR_VERSION >= 3
     if (PyUnicode_Check(obj)) {
         Py_ssize_t len;
         const char *ipstr = PyUnicode_AsUTF8AndSize(obj, &len);
-#else
-    if (PyUnicode_Check(obj) || PyString_Check(obj)) {
-        // Although this should work on Python 3, we will hopefully delete
-        // this soon and the Python 3 version is cleaner.
-        const char *ipstr = PyString_AsString(obj);
-        Py_ssize_t len = PyString_Size(obj);
-#endif
+
         if (!ipstr) {
             PyErr_SetString(PyExc_TypeError,
                             "argument 1 contains an invalid string");
@@ -270,14 +276,14 @@ static int ip_converter(PyObject *obj, struct sockaddr_storage *ip_address) {
         case 16: {
             ip_address->ss_family = AF_INET6;
             struct sockaddr_in6 *sin = (struct sockaddr_in6 *)ip_address;
-            memcpy(sin->sin6_addr.s6_addr, bytes, len);
+            memcpy(sin->sin6_addr.s6_addr, bytes, (size_t)len);
             Py_DECREF(packed);
             return 1;
         }
         case 4: {
             ip_address->ss_family = AF_INET;
             struct sockaddr_in *sin = (struct sockaddr_in *)ip_address;
-            memcpy(&(sin->sin_addr.s_addr), bytes, len);
+            memcpy(&(sin->sin_addr.s_addr), bytes, (size_t)len);
             Py_DECREF(packed);
             return 1;
         }
@@ -520,8 +526,13 @@ static PyObject *from_map(MMDB_entry_data_list_s **entry_data_list) {
         *entry_data_list = (*entry_data_list)->next;
 
         PyObject *key = PyUnicode_FromStringAndSize(
-            (char *)(*entry_data_list)->entry_data.utf8_string,
+            (*entry_data_list)->entry_data.utf8_string,
             (*entry_data_list)->entry_data.data_size);
+        if (!key) {
+            // PyUnicode_FromStringAndSize will set an appropriate exception
+            // in this case.
+            return NULL;
+        }
 
         *entry_data_list = (*entry_data_list)->next;
 
@@ -699,51 +710,46 @@ static PyTypeObject Metadata_Type = {
 
 static PyMethodDef MaxMindDB_methods[] = {{NULL, NULL, 0, NULL}};
 
-#if PY_MAJOR_VERSION >= 3
 static struct PyModuleDef MaxMindDB_module = {
     PyModuleDef_HEAD_INIT,
     .m_name = "extension",
     .m_doc = "This is a C extension to read MaxMind DB file format",
     .m_methods = MaxMindDB_methods,
 };
-#endif
 
-MOD_INIT(extension) {
+PyMODINIT_FUNC PyInit_extension(void) {
     PyObject *m;
 
-#if PY_MAJOR_VERSION >= 3
     m = PyModule_Create(&MaxMindDB_module);
-#else
-    m = Py_InitModule("extension", MaxMindDB_methods);
-#endif
 
     if (!m) {
-        RETURN_MOD_INIT(NULL);
+        return NULL;
     }
 
     Reader_Type.tp_new = PyType_GenericNew;
     if (PyType_Ready(&Reader_Type)) {
-        RETURN_MOD_INIT(NULL);
+        return NULL;
     }
     Py_INCREF(&Reader_Type);
     PyModule_AddObject(m, "Reader", (PyObject *)&Reader_Type);
 
     Metadata_Type.tp_new = PyType_GenericNew;
     if (PyType_Ready(&Metadata_Type)) {
-        RETURN_MOD_INIT(NULL);
+        return NULL;
     }
-    PyModule_AddObject(m, "extension", (PyObject *)&Metadata_Type);
+    Py_INCREF(&Metadata_Type);
+    PyModule_AddObject(m, "Metadata", (PyObject *)&Metadata_Type);
 
     PyObject *error_mod = PyImport_ImportModule("maxminddb.errors");
     if (error_mod == NULL) {
-        RETURN_MOD_INIT(NULL);
+        return NULL;
     }
 
     MaxMindDB_error = PyObject_GetAttrString(error_mod, "InvalidDatabaseError");
     Py_DECREF(error_mod);
 
     if (MaxMindDB_error == NULL) {
-        RETURN_MOD_INIT(NULL);
+        return NULL;
     }
 
     Py_INCREF(MaxMindDB_error);
@@ -751,5 +757,5 @@ MOD_INIT(extension) {
     /* We primarily add it to the module for backwards compatibility */
     PyModule_AddObject(m, "InvalidDatabaseError", MaxMindDB_error);
 
-    RETURN_MOD_INIT(m);
+    return m;
 }
